/*
---------------------------------------------------------------------------
This is a GAUSS program. It does empirical application for the paper of
Perron and Rodriguez (2016) "Residuals based Tests for Cointegration 
using GLS Detrending Data",
published in The Econometrics Journal, volume 19, pp. 84-111.
Written by:
Gabriel Rodríguez and Miguel Ataurima
Department of Economics
Pontificia Universidad Católica del Perú
June-2016
---------------------------------------------------------------------------
*/

@================================================================================@
@--------------------------- Table 1: PR (2016)----------------------------------@
@================================================================================@
/*
Table 1. Optimal value of c obtained from the MPGLS T test with R2 = 0.4
=========================================================================
m 		px = 0, py = 0 	px = 1, py= 1 	px = 1, py= 0
=========================================================================
x = 1 		−13.75 		−20.50 		−13.50
x = 2 		−18.25 		−23.75 		−18.00
x = 3 		−22.25 		−27.25 		−23.00
x = 4 		−26.25 		−30.75 		−26.00
x = 5 		−30.00 		−33.75 		−29.75
=========================================================================
*/    

library pgraph;	                                             @open library "pgraph" in case we need to graph the time series@

/*
Loading the data. First column is the variable Y. The other columns are 
the variables X. In the example below, we load data of T=95 and 
5 variables. 
First variable is Y: Peruvian Consumption. Other variables are X: 
Peruvian Private Investment, Peruvian Public Investment, Peruvian GDP, 
and Terms of Trade.
*/

load data[95,5]=datos.txt;                                  @load data: y and x@

y=data[.,1];                                                @setting data for dependent variable Y@
x=data[.,2:5];                                              @setting data for variables X@

m=cols(x);                                                  @m=number of I(1) variables in X. Using m=cols(x) assumes all variable are I(1), the usual case. Otherwise set m to the desired value@                                          
bigt=rows(y);                                               @sample size@
det_comp=1;                                                 @type of deterministic components:@
                                                            @ 1 = constant in regression and non-trending data: py=0, px=0@
                                                            @ 2 = constant and time trend in the regression (with or without trending data): py=1, px=1@
                                                            @ 3 = constant only in the regression with trending data (deterministic cointegration): py=0, px=1@
@__________________________________________________________________________@
@ The code below should not be modified. If you spot errors, please contact@
@ Gabriel Rodríguez (gabriel.rodriguez(at)pucp.edu.pe)@
@__________________________________________________________________________@

@"optimal CBAR" (opt_cbar) selected based on m and det_comp using Table 1 of PR (2016)@
@ You may also see the Table 1 of PR (2016) at the beginning of this code@

if m==1;
    options = {-13.75, -20.50, -13.50};
elseif m==2;
    options = {-18.25, -23.75, -18.00};
elseif m==3;
    options = {-22.25, -27.25, -23.00};
elseif m==4;
     options = {-26.25, -30.75, -26.00};
else;
     options = {-30.00, -33.75, -29.75};
endif;
opt_cbar = options[det_comp];


kmin=0;                                                          @minimum lag length@
kmax=round(4*(bigt/100)^(1/4));                                  @maximum lag length@

y_d=glsdy(y,bigt,opt_cbar,det_comp);                             @GLS detrending of variable Y@
x_d=glsdx(x,bigt,opt_cbar,det_comp,m);                           @GLS detrending of variables X@

{beta_hat,ud_hat,fit_hat}=olsqr2(y_d,x_d);                       @equation of cointegration. ud_hat are the residuals to construct the tests@
           
{gls_tests,kbic}=tests(ud_hat,kmin,kmax,opt_cbar,det_comp);      @construction of the GLS tests. Lag length (k) is selected by BIC@

/*
Print of the Results. Please, see Tables 2,3,4 of PR (2016) for correspondig
critical values:
Table 2: if det_comp==1 (py=0, px=0);
Table 3: if det_comp==2 (py=1, px=1);
Table 4: if det_comp==3 (py=0, px=1).
*/

print "======================================================================";
print "--------------- THE GLS TESTS FOR COINTEGRATION ----------------------";
print "======================================================================";

print "number of X variables="             m;
print "optimal cbar="                      opt_cbar;
print "type of deterministic components="  det_comp;
print "lag="                               kbic;
print "======================================================================";
print "MZ_rho="  gls_tests[1,1];
print "MSB=   "  gls_tests[2,1];
print "MZ_t=  "  gls_tests[3,1];
print "ADF=   "  gls_tests[4,1];
print "Z_rho= "  gls_tests[5,1];
print "Z_t_rho=" gls_tests[6,1];
print "MPT=   "  gls_tests[7,1];
print "======================================================================";

@================================================================================@
@-------------------------------- procedures ------------------------------------@
@================================================================================@
/*
procedure GLSY: performs GLS detrending of the y variable.
inputs:
y=left-hand side variable in the cointegrating equation
bigt=sample size
opt_cbar=value of the parameter CBAR where GLS detrending is performed. 
         The value depends of the number of Xs variables and the type of the 
         deterministic components. See Table 1 of PR (2016)
*/

proc(1)=glsdy(y,bigt,opt_cbar,det_comp);               @GLS DETRENDING PROC. for Y@
local z,t,alphabar,zalpha,ztalpha,yalpha,psi_y,y_d,talpha;

if det_comp==1;

z=ones(bigt,1);
alphabar=1+(opt_cbar/bigt);

yalpha=zeros(bigt,1);
zalpha=zeros(bigt,1);

yalpha[1:1,.]=y[1:1,.];
zalpha[1:1,.]=z[1:1,.];
yalpha[2:bigt,.]=y[2:bigt,.]-alphabar*y[1:bigt-1,.];
zalpha[2:bigt,.]=z[2:bigt,.]-alphabar*z[1:bigt-1,.];

psi_y=inv(zalpha'zalpha)*(zalpha'yalpha);
y_d=y-z*psi_y;

endif;


if det_comp==2;

z=ones(bigt,1);
t=seqa(1,1,bigt);
alphabar=1+(opt_cbar/bigt);

yalpha=zeros(bigt,1);
zalpha=zeros(bigt,1);
talpha=zeros(bigt,1);

yalpha[1:1,.]=y[1:1,.];
zalpha[1:1,.]=z[1:1,.];
talpha[1:1,.]=t[1:1,.];
yalpha[2:bigt,.]=y[2:bigt,.]-alphabar*y[1:bigt-1,.];
zalpha[2:bigt,.]=z[2:bigt,.]-alphabar*z[1:bigt-1,.];
talpha[2:bigt,.]=t[2:bigt,.]-alphabar*t[1:bigt-1,.];

ztalpha=zalpha~talpha;

psi_y=inv(ztalpha'ztalpha)*(ztalpha'yalpha);
y_d=y-(z~t)*psi_y;

endif;

if det_comp==3;

z=ones(bigt,1);
alphabar=1+(opt_cbar/bigt);

yalpha=zeros(bigt,1);
zalpha=zeros(bigt,1);

yalpha[1:1,.]=y[1:1,.];
zalpha[1:1,.]=z[1:1,.];
yalpha[2:bigt,.]=y[2:bigt,.]-alphabar*y[1:bigt-1,.];
zalpha[2:bigt,.]=z[2:bigt,.]-alphabar*z[1:bigt-1,.];

psi_y=inv(zalpha'zalpha)*(zalpha'yalpha);
y_d=y-z*psi_y;

endif;

retp(y_d);
endp;

@================================================================================@
@-------------------------------- procedures ------------------------------------@
@================================================================================@
/*
procedure GLSX: performs GLS detrending of the X variables.
inputs:
x=rigth-hand side variables in the cointegrating equation
bigt=sample size
opt_cbar=value of the parameter CBAR where GLS detrending is performed. 
         The value depends of the number of Xs variables and the type of the 
         deterministic components. See Table 1 of PR (2016)
m=number of X variables
*/

proc(1)=glsdx(x,bigt,opt_cbar,det_comp,m);               @GLS DETRENDING PROC. for X's@
local z,t,alphabar,zalpha,talpha,xalpha,psi_x,x_d,i,ztalpha;

if det_comp==1;

z=ones(bigt,1);
alphabar=1+(opt_cbar/bigt);

xalpha=zeros(bigt,m);
zalpha=zeros(bigt,1);
x_d=zeros(bigt,m);

zalpha[1:1,.]=z[1:1,.];
zalpha[2:bigt,.]=z[2:bigt,.]-alphabar*z[1:bigt-1,.];

i=1;
do while i<= m;

xalpha[1:1,i]=x[1:1,i];
xalpha[2:bigt,i]=x[2:bigt,i]-alphabar*x[1:bigt-1,i];

psi_x=inv(zalpha'zalpha)*(zalpha'xalpha[.,i]);
x_d[.,i]=x[.,i]-z*psi_x;

i=i+1;
endo;

endif;

if det_comp==2;

z=ones(bigt,1);
t=seqa(1,1,bigt);
alphabar=1+(opt_cbar/bigt);

xalpha=zeros(bigt,m);
zalpha=zeros(bigt,1);
talpha=zeros(bigt,1);
x_d=zeros(bigt,m);

zalpha[1:1,.]=z[1:1,.];
talpha[1:1,.]=t[1:1,.];
zalpha[2:bigt,.]=z[2:bigt,.]-alphabar*z[1:bigt-1,.];
talpha[2:bigt,.]=t[2:bigt,.]-alphabar*t[1:bigt-1,.];

i=1;
do while i<=m;

xalpha[1:1,i]=x[1:1,i];
xalpha[2:bigt,i]=x[2:bigt,i]-alphabar*x[1:bigt-1,i];

ztalpha=zalpha~talpha;

psi_x=inv(ztalpha'ztalpha)*(ztalpha'xalpha[.,i]);
x_d[.,i]=x[.,i]-(z~t)*psi_x;

i=i+1;
endo;

endif;

if det_comp==3;

z=ones(bigt,1);
alphabar=1+(opt_cbar/bigt);

xalpha=zeros(bigt,m);
zalpha=zeros(bigt,1);
x_d=zeros(bigt,m);

zalpha[1:1,.]=z[1:1,.];
zalpha[2:bigt,.]=z[2:bigt,.]-alphabar*z[1:bigt-1,.];

i=1;
do while i<= m;

xalpha[1:1,i]=x[1:1,i];
xalpha[2:bigt,i]=x[2:bigt,i]-alphabar*x[1:bigt-1,i];

psi_x=inv(zalpha'zalpha)*(zalpha'xalpha[.,i]);
x_d[.,i]=x[.,i]-z*psi_x;

i=i+1;
endo;

endif;

retp(x_d[.,.]);
endp;


@================================================================================@
@-------------------------------- procedures ------------------------------------@
@================================================================================@
/*
procedure TESTS: construct the GLS tests.
inputs:
ud_hat: residuals from the cointegrating equation
kmin=minimun lag length
kmax=maximum lag length
opt_cbar=value of the parameter CBAR where GLS detrending is performed. 
         The value depends of the number of Xs variables and the type of the 
         deterministic components. See Table 1 of PR (2016)
*/

proc(2)=tests(ud_hat,kmin,kmax,opt_cbar,det_comp);
local gls_tests,aa,ss,talpha,sum_ud,lag_bic,dud_hat,reg,h,dud_hat_1,reg_1,sumb,nef2,s2e2,xx2,s22,bicgls,minbicgls,ud_hat_1,sum_ud_hat,b2,j,alpha_1,ee_1,e2,fit_1,fit2,kbic;

if det_comp==1;
	
gls_tests=zeros(7,1);
minbicgls=999999999;

{alpha_1,ee_1,fit_1}=olsqr2(ud_hat[2:rows(ud_hat),1],ud_hat[1:rows(ud_hat)-1,1]);
ud_hat_1=ud_hat[1:rows(ud_hat)-1,1];

aa=inv(ud_hat_1'ud_hat_1);
ss=ee_1'ee_1/(rows(ee_1)-cols(ud_hat_1));

talpha=(alpha_1-1)/sqrt(ss*aa[1,1]);
sum_ud_hat=sumc(ud_hat[1:bigt-1,1]^2)/(bigt-1)^2;

lag_bic=kmin;                                      @loop for lag@
do while lag_bic<=kmax;

dud_hat=diff(ud_hat,1);
reg=lagn(ud_hat,1);

h=1;
do while h<=lag_bic;
reg=reg~lagn(dud_hat,h);
h=h+1;
endo;

dud_hat_1=trimr(dud_hat,lag_bic+1,0);
reg_1=trimr(reg,lag_bic+1,0);

{b2,e2,fit2}=olsqr2(dud_hat_1,reg_1);
nef2=rows(e2);
s2e2=e2'e2/(nef2-cols(reg_1));
xx2=inv(reg_1'reg_1);

@loop for sum of coeff. "b"@

sumb=0; 
 j=1;
 do while j<=lag_bic;
 sumb=sumb+b2[j+1,1];
 j=j+1;
 endo;

s22=s2e2/((1-sumb)^2);  @calculate "spectral density estimator (S2)"@

bicgls=ln(e2'e2/(bigt-kmax))+(ln(bigt-kmax)*(lag_bic))/(bigt-kmax); @calc. bic,search minimun of bic@
if bicgls .< minbicgls;
  minbicgls=bicgls;
  kbic=lag_bic;

gls_tests[1,1]=((ud_hat[bigt,1]^2/bigt)-s22)*((2*sum_ud_hat))^(-1);   
gls_tests[2,1]=sqrt(sum_ud_hat/s22);
gls_tests[3,1]=gls_tests[1,1]*gls_tests[2,1];
gls_tests[4,1]=b2[1,1]/sqrt(s2e2*xx2[1,1]);
gls_tests[5,1]=(bigt-1)*(alpha_1-1)-0.5*(s22-ss)*(sum_ud_hat)^(-1);
gls_tests[6,1]=sqrt(ss/s22)*talpha-(s22-ss)/sqrt(4*s22*sum_ud_hat);
gls_tests[7,1]=( (opt_cbar^2)*(sum_ud_hat)-(opt_cbar*(ud_hat[bigt,1]^2/bigt)))/s22;

endif;

lag_bic=lag_bic+1;
endo;

endif;

if det_comp==2;
	
gls_tests=zeros(7,1);
minbicgls=999999999;

{alpha_1,ee_1,fit_1}=olsqr2(ud_hat[2:rows(ud_hat),1],ud_hat[1:rows(ud_hat)-1,1]);
ud_hat_1=ud_hat[1:rows(ud_hat)-1,1];

aa=inv(ud_hat_1'ud_hat_1);
ss=ee_1'ee_1/(rows(ee_1)-cols(ud_hat_1));

talpha=(alpha_1-1)/sqrt(ss*aa[1,1]);
sum_ud_hat=sumc(ud_hat[1:bigt-1,1]^2)/(bigt-1)^2;

lag_bic=kmin;                                      @loop for lag@
do while lag_bic<=kmax;

dud_hat=diff(ud_hat,1);
reg=lagn(ud_hat,1);

h=1;
do while h<=lag_bic;
reg=reg~lagn(dud_hat,h);
h=h+1;
endo;

dud_hat_1=trimr(dud_hat,lag_bic+1,0);
reg_1=trimr(reg,lag_bic+1,0);

{b2,e2,fit2}=olsqr2(dud_hat_1,reg_1);
nef2=rows(e2);
s2e2=e2'e2/(nef2-cols(reg_1));
xx2=inv(reg_1'reg_1);

@loop for sum of coeff. "b"@

sumb=0; 
 j=1;
 do while j<=lag_bic;
 sumb=sumb+b2[j+1,1];
 j=j+1;
 endo;

s22=s2e2/((1-sumb)^2);  @calculate "spectral density estimator (S2)"@

bicgls=ln(e2'e2/(bigt-kmax))+(ln(bigt-kmax)*(lag_bic))/(bigt-kmax); @calc. bic,search minimun of bic@
if bicgls .< minbicgls;
  minbicgls=bicgls;
  kbic=lag_bic;

gls_tests[1,1]=((ud_hat[bigt,1]^2/bigt)-s22)*((2*sum_ud_hat))^(-1);   
gls_tests[2,1]=sqrt(sum_ud_hat/s22);
gls_tests[3,1]=gls_tests[1,1]*gls_tests[2,1];
gls_tests[4,1]=b2[1,1]/sqrt(s2e2*xx2[1,1]);
gls_tests[5,1]=(bigt-1)*(alpha_1-1)-0.5*(s22-ss)*(sum_ud_hat)^(-1);
gls_tests[6,1]=sqrt(ss/s22)*talpha-(s22-ss)/sqrt(4*s22*sum_ud_hat);
gls_tests[7,1]=( (opt_cbar^2)*(sum_ud_hat)+(1-opt_cbar)*(ud_hat[bigt,1]^2/bigt))/s22;

endif;

lag_bic=lag_bic+1;
endo;

endif;

if det_comp==3;
	
gls_tests=zeros(7,1);
minbicgls=999999999;

{alpha_1,ee_1,fit_1}=olsqr2(ud_hat[2:rows(ud_hat),1],ud_hat[1:rows(ud_hat)-1,1]);
ud_hat_1=ud_hat[1:rows(ud_hat)-1,1];

aa=inv(ud_hat_1'ud_hat_1);
ss=ee_1'ee_1/(rows(ee_1)-cols(ud_hat_1));

talpha=(alpha_1-1)/sqrt(ss*aa[1,1]);
sum_ud_hat=sumc(ud_hat[1:bigt-1,1]^2)/(bigt-1)^2;

lag_bic=kmin;                                      @loop for lag@
do while lag_bic<=kmax;

dud_hat=diff(ud_hat,1);
reg=lagn(ud_hat,1);

h=1;
do while h<=lag_bic;
reg=reg~lagn(dud_hat,h);
h=h+1;
endo;

dud_hat_1=trimr(dud_hat,lag_bic+1,0);
reg_1=trimr(reg,lag_bic+1,0);

{b2,e2,fit2}=olsqr2(dud_hat_1,reg_1);
nef2=rows(e2);
s2e2=e2'e2/(nef2-cols(reg_1));
xx2=inv(reg_1'reg_1);

@loop for sum of coeff. "b"@

sumb=0; 
 j=1;
 do while j<=lag_bic;
 sumb=sumb+b2[j+1,1];
 j=j+1;
 endo;

s22=s2e2/((1-sumb)^2);  @calculate "spectral density estimator (S2)"@

bicgls=ln(e2'e2/(bigt-kmax))+(ln(bigt-kmax)*(lag_bic))/(bigt-kmax); @calc. bic,search minimun of bic@
if bicgls .< minbicgls;
  minbicgls=bicgls;
  kbic=lag_bic;

gls_tests[1,1]=((ud_hat[bigt,1]^2/bigt)-s22)*((2*sum_ud_hat))^(-1);   
gls_tests[2,1]=sqrt(sum_ud_hat/s22);
gls_tests[3,1]=gls_tests[1,1]*gls_tests[2,1];
gls_tests[4,1]=b2[1,1]/sqrt(s2e2*xx2[1,1]);
gls_tests[5,1]=(bigt-1)*(alpha_1-1)-0.5*(s22-ss)*(sum_ud_hat)^(-1);
gls_tests[6,1]=sqrt(ss/s22)*talpha-(s22-ss)/sqrt(4*s22*sum_ud_hat);
gls_tests[7,1]=( (opt_cbar^2)*(sum_ud_hat)-(opt_cbar*(ud_hat[bigt,1]^2/bigt)))/s22;

endif;

lag_bic=lag_bic+1;
endo;

endif;

retp(gls_tests,kbic);
endp;


@================================================================================@
@-------------------------------- procedures ------------------------------------@
@================================================================================@

proc(1)=diff(x,k);    @procedure to generate first differences of the series@
  if (k == 0);
  retp(x);

  endif;
  retp(zeros(k,cols(x))|trimr(x,k,0)-trimr(lagn(x,k),k,0));
endp;

proc(1)=lagn(x,n);     @procedure to generate teh first lag of the series@
   local f;
   if n > 0;
   f = zeros(n,cols(x))|trimr(x,0,n);
   else;
   f = trimr(x,abs(n),0)|zeros(abs(n),cols(x));
   endif;
   retp(f);
endp;

