function code_gls_cointegration

% ---------------------------------------------------------------------------
% This is a MATLAB program. It does empirical application for the paper of
% Perron and Rodriguez (2016) "Residuals based Tests for Cointegration 
% using GLS Detrending Data",
% published in The Econometrics Journal, volume 19, pp. 84-111.
% Written by:
% Gabriel Rodrguez and Miguel Ataurima
% Department of Economics
% Pontificia Universidad Catlica del Per
% June-2016
% ---------------------------------------------------------------------------

% Loadind the data. First column is the variable Y. The other columns are 
% the variables X. In the example below, we load data of T=95 and 
% 5 variables. 
% First variable is Y: Peruvian Consumption.Other variables are X: 
% Peruvian Private Investment, Peruvian Public Investment, Peruvian GDP, 
% and Terms of Trade.

% =========================================================================
% --------------------------- Table 1: PR (2016)---------------------------
% =========================================================================
% 
% Table 1. Optimal value of c obtained from the MPGLS T test with R2 = 0.4
% =========================================================================
% m 		px = 0, py = 0 	px = 1, py= 1 	px = 1, py= 0
% =========================================================================
% x = 1 		-13.75 		-20.50 		-13.50
% x = 2 		-18.25 		-23.75 		-18.00
% x = 3 		-22.25 		-27.25 		-23.00
% x = 4 		-26.25 		-30.75 		-26.00
% x = 5 		-30.00 		-33.75 		-29.75
% =========================================================================

fprintf('\n');
data = load('datos.txt');
y = data(:,1);            % setting data for dependent variable Y
x = data(:,2:5);          % setting data for variables X

m = cols(x);              % m=number of I(1) variables in X. Using m=cols(x) assumes all variable are I(1), the usual case. Otherwise set m to the desired value; 
bigt = rows(y);           % sample size
det_comp = 1;             % type of deterministic components: 
                          %   1 = constant in regression and non-trending data: py=0, px=0;  
                          %   2 = constant and time trend in the regression (with or without trending data): py=1, px=1; 
                          %   3 = constant only in the regression with trending data (deterministic cointegration): py=0, px=1
%__________________________________________________________________________
% The code below should not be modified. If you spot errors, please contact
% Gabriel Rodrguez (gabriel.rodriguez@pucp.edu.pe)
%__________________________________________________________________________

% "optimal CBAR (opt_cbar) selected based on m and det_comp using Table 1 of PR (2016)

switch m
    case 1
        options = [-13.75 			-20.50 			-13.50];
    case 2
        options = [-18.25 			-23.75 			-18.00];
    case 3
        options = [-22.25 			-27.25 			-23.00];
    case 4
        options = [-26.25 			-30.75 			-26.00];
    case 5
        options = [-30.00 			-33.75 			-29.75];
end

% selection of opt_cbar
opt_cbar = options(det_comp);

kmin = 0;                                 % minimum lag length@
kmax = round(4*(bigt/100)^(1/4));         % maximum lag length@

y_d = glsdy(y,bigt,opt_cbar,det_comp);    % GLS detrending of variable Y
x_d = glsdx(x,bigt,opt_cbar,det_comp,m);  % GLS detrending of variables X

[~, ud_hat, ~] = olsqr2(y_d,x_d);                                    % equation of cointegration. ud_hat are the residuals to construct the tests
           
[gls_tests,kbic] = tests(ud_hat,kmin,kmax,opt_cbar,det_comp);        % construction of the GLS tests. Lag length (k) is selected by BIC

% Print of the Results. Please, see Tables 2,3,4 of PR (2016) for correspondig
% critical values:
% Table 2: if det_comp==1 (py=0, px=0);
% Table 3: if det_comp==2 (py=1, px=1);
% Table 4: if det_comp==3 (py=0, px=1).


fprintf('======================================================================\n');
fprintf('--------------- THE GLS TESTS FOR COINTEGRATION ----------------------\n');
fprintf('======================================================================\n');

fprintf('number of X variables            = %8.3f\n', m);
fprintf('optimal cbar                     = %8.3f\n', opt_cbar);
fprintf('type of deterministic components = %8.3f\n', det_comp);
fprintf('lag                              = %8.3f\n', kbic);
fprintf('======================================================================\n');
fprintf('MZ_rho  = %12.6f\n', gls_tests(1,1));
fprintf('MSB     = %12.6f\n', gls_tests(2,1));
fprintf('MZ_t    = %12.6f\n', gls_tests(3,1));
fprintf('ADF     = %12.6f\n', gls_tests(4,1));
fprintf('Z_rho   = %12.6f\n', gls_tests(5,1));
fprintf('Z_t_rho = %12.6f\n', gls_tests(6,1));
fprintf('MPT     = %12.6f\n', gls_tests(7,1));
fprintf('======================================================================\n');


% ================================================================================
% -------------------------------- procedures ------------------------------------
% ================================================================================

% procedure GLSY: performs GLS detrending of the y variable.
% inputs:
% y=left-hand side variable in the cointegrating equation
% bigt=sample size
% opt_cbar=value of the parameter CBAR where GLS detrending is performed.The value
% depends of the number of Xs variables and the type of the deterministic 
% components. See Table 1 of PR (2016)


function y_d = glsdy(y,bigt,opt_cbar,det_comp)               % GLS DETRENDING PROC. for Y

    if det_comp==1

        z = ones(bigt,1);
        alphabar = 1 + opt_cbar/bigt;

        yalpha = zeros(bigt,1);
        zalpha = zeros(bigt,1);

        yalpha(1:1,:) = y(1:1,:);
        zalpha(1:1,:) = z(1:1,:);
        yalpha(2:bigt,:) = y(2:bigt,:) - alphabar*y(1:bigt-1,:);
        zalpha(2:bigt,:) = z(2:bigt,:) - alphabar*z(1:bigt-1,:);

        psi_y = (zalpha'*zalpha)^-1*(zalpha'*yalpha);
        y_d = y-z*psi_y;

    elseif det_comp==2

        z = ones(bigt,1);
        t = seqa(1,1,bigt);
        alphabar = 1 + opt_cbar/bigt;

        yalpha = zeros(bigt,1);
        zalpha = zeros(bigt,1);
        talpha = zeros(bigt,1);

        yalpha(1:1,:) = y(1:1,:);
        zalpha(1:1,:) = z(1:1,:);
        talpha(1:1,:) = t(1:1,:);
        yalpha(2:bigt,:) = y(2:bigt,:) - alphabar*y(1:bigt-1,:);
        zalpha(2:bigt,:) = z(2:bigt,:) - alphabar*z(1:bigt-1,:);
        talpha(2:bigt,:) = t(2:bigt,:) - alphabar*t(1:bigt-1,:);

        ztalpha = [zalpha talpha];

        psi_y = (ztalpha'*ztalpha)^-1*(ztalpha'*yalpha);
        y_d = y-[z t]*psi_y;

    elseif det_comp==3

        z = ones(bigt,1);
        alphabar = 1 + opt_cbar/bigt;

        yalpha = zeros(bigt,1);
        zalpha = zeros(bigt,1);

        yalpha(1:1,:) = y(1:1,:);
        zalpha(1:1,:) = z(1:1,:);
        yalpha(2:bigt,:) = y(2:bigt,:) - alphabar*y(1:bigt-1,:);
        zalpha(2:bigt,:) = z(2:bigt,:) - alphabar*z(1:bigt-1,:);

        psi_y = (zalpha'*zalpha)^-1*(zalpha'*yalpha);
        y_d = y-z*psi_y;
    end

end

% ================================================================================
% -------------------------------- procedures ------------------------------------
% ================================================================================

% procedure GLSX: performs GLS detrending of the X variables.
% inputs:
% x=rigth-hand side variables in the cointegrating equation
% bigt=sample size
% opt_cbar=value of the parameter CBAR where GLS detrending is performed.The value
% depends of the number of Xs variables and the type of the deterministic 
% components. See Table 1 of PR (2016)
% m=number of X variables

function argout = glsdx(x,bigt,opt_cbar,det_comp,m)               %GLS DETRENDING PROC. for X's

    if det_comp==1

        z = ones(bigt,1);
        alphabar = 1 + opt_cbar/bigt;

        xalpha = zeros(bigt,m);
        zalpha = zeros(bigt,1);
        x_d = zeros(bigt,m);

        zalpha(1:1,:) = z(1:1,:);
        zalpha(2:bigt,:) = z(2:bigt,:) - alphabar*z(1:bigt-1,:);

        for i = 1:m
            xalpha(1:1,i) = x(1:1,i);
            xalpha(2:bigt,i) = x(2:bigt,i) - alphabar*x(1:bigt-1,i);

            psi_x = (zalpha'*zalpha)^-1*(zalpha'*xalpha(:,i));
            x_d(:,i) = x(:,i) - z*psi_x;
        end

    elseif det_comp==2

        z=ones(bigt,1);
        t=seqa(1,1,bigt);
        alphabar=1+(opt_cbar/bigt);

        xalpha=zeros(bigt,m);
        zalpha=zeros(bigt,1);
        talpha=zeros(bigt,1);
        x_d=zeros(bigt,m);

        zalpha(1:1,:) = z(1:1,:);
        talpha(1:1,:) = t(1:1,:);
        zalpha(2:bigt,:) = z(2:bigt,:) - alphabar*z(1:bigt-1,:);
        talpha(2:bigt,:) = t(2:bigt,:) - alphabar*t(1:bigt-1,:);

        for i = 1:m
            
            xalpha(1:1,i) = x(1:1,i);
            xalpha(2:bigt,i) = x(2:bigt,i)-alphabar*x(1:bigt-1,i);

            ztalpha = [zalpha talpha];

            psi_x = (ztalpha'*ztalpha)^-1*(ztalpha'*xalpha(:,i));
            x_d(:,i) = x(:,i) - [z t]*psi_x;
        end
        
    elseif det_comp==3

         z = ones(bigt,1);
        alphabar = 1 + opt_cbar/bigt;

        xalpha = zeros(bigt,m);
        zalpha = zeros(bigt,1);
        x_d = zeros(bigt,m);

        zalpha(1:1,:) = z(1:1,:);
        zalpha(2:bigt,:) = z(2:bigt,:) - alphabar*z(1:bigt-1,:);

        for i = 1:m
            xalpha(1:1,i) = x(1:1,i);
            xalpha(2:bigt,i) = x(2:bigt,i) - alphabar*x(1:bigt-1,i);

            psi_x = (zalpha'*zalpha)^-1*(zalpha'*xalpha(:,i));
            x_d(:,i) = x(:,i) - z*psi_x;
        end

    end
    argout = x_d(:,:);
end


% ================================================================================
% -------------------------------- procedures ------------------------------------
% ================================================================================

% procedure TESTS: construct the GLS tests.
% inputs:
% ud_hat: residuals from the cointegrating equation
% kmin=minimun lag length
% kmax=maximum lag length
% opt_cbar=value of the parameter CBAR where GLS detrending is performed.The value
% depends of the number of Xs variables and the type of the deterministic 
% components. See Table 1 of PR (2016)

function [gls_tests,kbic] = tests(ud_hat,kmin,kmax,opt_cbar,det_comp)

    if det_comp == 1

        gls_tests=zeros(7,1);
        minbicgls = 999999999;

        [alpha_1,ee_1,~] = olsqr2(ud_hat(2:rows(ud_hat),1),ud_hat(1:rows(ud_hat)-1,1));
        ud_hat_1 = ud_hat(1:rows(ud_hat)-1,1);

        aa = inv(ud_hat_1'*ud_hat_1);
        ss = ee_1'*ee_1/(rows(ee_1)-cols(ud_hat_1));

        talpha = (alpha_1-1)/sqrt(ss*aa(1,1));
        sum_ud_hat = sum(ud_hat(1:bigt-1,1).^2)/(bigt-1)^2;

        % loop for lag
        for lag_bic = kmin:kmax

            dud_hat = diff(ud_hat,1);
            
            reg = lagn(ud_hat,1);

            for h = 1:lag_bic
                reg = [reg lagn(dud_hat,h)];
            end

            dud_hat_1 = trimr(dud_hat,lag_bic+1,0);
            reg_1 = trimr(reg,lag_bic+1,0);

            [b2,e2,~] = olsqr2(dud_hat_1,reg_1);
            nef2 = rows(e2);
            s2e2 = e2'*e2/(nef2-cols(reg_1));
            xx2 = inv(reg_1'*reg_1);

            % loop for sum of coeff. ',b',@
            sumb = 0; 
            for j = 1:lag_bic
                sumb = sumb + b2(j+1,1);
            end
            
            % calculate spectral density estimator (S2),
            s22 = s2e2/((1-sumb)^2);

            bicgls = log(e2'*e2/(bigt-kmax)) + (log(bigt-kmax)*(lag_bic))/(bigt-kmax); %calc. bic,search minimun of bic@
            if bicgls < minbicgls
                minbicgls=bicgls;
                kbic=lag_bic;

                gls_tests(1,1)=((ud_hat(bigt,1)^2/bigt)-s22)*((2*sum_ud_hat))^(-1);   
                gls_tests(2,1)=sqrt(sum_ud_hat/s22);
                gls_tests(3,1)=gls_tests(1,1)*gls_tests(2,1);
                gls_tests(4,1)=b2(1,1)/sqrt(s2e2*xx2(1,1));
                gls_tests(5,1)=(bigt-1)*(alpha_1-1)-0.5*(s22-ss)*(sum_ud_hat)^(-1);
                gls_tests(6,1)=sqrt(ss/s22)*talpha-(s22-ss)/sqrt(4*s22*sum_ud_hat);
                gls_tests(7,1)=( (opt_cbar^2)*(sum_ud_hat)-(opt_cbar*(ud_hat(bigt,1)^2/bigt)))/s22;
            end

        end
        
    elseif det_comp == 2

        gls_tests = zeros(7,1);
        minbicgls = 999999999;

        [alpha_1,ee_1,~] = olsqr2(ud_hat(2:rows(ud_hat),1),ud_hat(1:rows(ud_hat)-1,1));
        ud_hat_1 = ud_hat(1:rows(ud_hat)-1,1);

        aa = inv(ud_hat_1'*ud_hat_1);
        ss = ee_1'*ee_1/(rows(ee_1)-cols(ud_hat_1));

        talpha = (alpha_1-1)/sqrt(ss*aa(1,1));
        sum_ud_hat = sum(ud_hat(1:bigt-1,1).^2)/(bigt-1)^2;

        % loop for lag
        for lag_bic = kmin:kmax

            dud_hat = diff(ud_hat,1);
            reg = lagn(ud_hat,1);

            for h = 1:lag_bic
                reg = [reg lagn(dud_hat,h)];
            end

            dud_hat_1 = trimr(dud_hat,lag_bic+1,0);
            reg_1 = trimr(reg,lag_bic+1,0);

            [b2,e2,~] = olsqr2(dud_hat_1,reg_1);
            nef2 = rows(e2);
            s2e2 = e2'*e2/(nef2-cols(reg_1));
            xx2 = inv(reg_1'*reg_1);

            % loop for sum of coeff. ',b',
            sumb = 0; 
            for j = 1:lag_bic
                sumb = sumb+b2(j+1,1);
            end
            
            % calculate the spectral density estimator (S2),
            s22 = s2e2/((1-sumb)^2);

            bicgls = log(e2'*e2/(bigt-kmax)) + (log(bigt-kmax)*(lag_bic))/(bigt-kmax); %calc. bic,search minimun of bic@
            if bicgls < minbicgls
                minbicgls = bicgls;
                kbic = lag_bic;

                gls_tests(1,1) = ((ud_hat(bigt,1)^2/bigt)-s22)*((2*sum_ud_hat))^(-1);   
                gls_tests(2,1) = sqrt(sum_ud_hat/s22);
                gls_tests(3,1) = gls_tests(1,1)*gls_tests(2,1);
                gls_tests(4,1) = b2(1,1)/sqrt(s2e2*xx2(1,1));
                gls_tests(5,1) = (bigt-1)*(alpha_1-1)-0.5*(s22-ss)*(sum_ud_hat)^(-1);
                gls_tests(6,1) = sqrt(ss/s22)*talpha-(s22-ss)/sqrt(4*s22*sum_ud_hat);
                gls_tests(7,1) = ( (opt_cbar^2)*(sum_ud_hat)+(1-opt_cbar)*(ud_hat(bigt,1)^2/bigt))/s22;

            end
        end

    elseif det_comp==3

        gls_tests = zeros(7,1);
        minbicgls = 999999999;

        [alpha_1,ee_1,~] = olsqr2(ud_hat(2:rows(ud_hat),1),ud_hat(1:rows(ud_hat)-1,1));
        ud_hat_1 = ud_hat(1:rows(ud_hat)-1,1);

        aa = inv(ud_hat_1'*ud_hat_1);
        ss = ee_1'*ee_1/(rows(ee_1)-cols(ud_hat_1));

        talpha = (alpha_1-1)/sqrt(ss*aa(1,1));
        sum_ud_hat = sum(ud_hat(1:bigt-1,1).^2)/(bigt-1)^2;

        % loop for lag
        for lag_bic = kmin:kmax

            dud_hat = diff(ud_hat,1);
            reg = lagn(ud_hat,1);

            for h = 1:lag_bic
                reg = [reg lagn(dud_hat,h)];
            end

            dud_hat_1 = trimr(dud_hat,lag_bic+1,0);
            reg_1 = trimr(reg,lag_bic+1,0);

            [b2,e2,~] = olsqr2(dud_hat_1,reg_1);
            nef2 = rows(e2);
            s2e2 = e2'*e2/(nef2-cols(reg_1));
            xx2 = inv(reg_1'*reg_1);

            % loop for sum of coeff. ',b',
            sumb = 0; 
            for j = 1:lag_bic
                sumb = sumb + b2(j+1,1);
            end
            
            % calculate the spectral density estimator (S2),
            s22 = s2e2/((1-sumb)^2);

            bicgls = log(e2'*e2/(bigt-kmax)) + (log(bigt-kmax)*(lag_bic))/(bigt-kmax); %calc. bic,search minimun of bic@
            if bicgls < minbicgls
                minbicgls = bicgls;
                kbic = lag_bic;

                gls_tests(1,1) = ((ud_hat(bigt,1)^2/bigt)-s22)*((2*sum_ud_hat))^(-1);   
                gls_tests(2,1) = sqrt(sum_ud_hat/s22);
                gls_tests(3,1) = gls_tests(1,1)*gls_tests(2,1);
                gls_tests(4,1) = b2(1,1)/sqrt(s2e2*xx2(1,1));
                gls_tests(5,1) = (bigt-1)*(alpha_1-1)-0.5*(s22-ss)*(sum_ud_hat)^(-1);
                gls_tests(6,1) = sqrt(ss/s22)*talpha-(s22-ss)/sqrt(4*s22*sum_ud_hat);
                gls_tests(7,1)=( (opt_cbar^2)*(sum_ud_hat)-(opt_cbar*(ud_hat(bigt,1)^2/bigt)))/s22;

            end
        end

    end
end
    

end

% ================================================================================
% -------------------------------- procedures ------------------------------------
% ================================================================================

function z = diff(x,k)    
% procedure to generate first differences of the series
    if k == 0
        z = x;
    else
        z = [zeros(k,cols(x)); trimr(x,k,0)-trimr(lagn(x,k),k,0)];
    end
end

function f = lagn(x,n)
% procedure to generate one lag of the series
    if n > 0
        f = [zeros(n,cols(x)); trimr(x,0,n)];
    else
        f = [trimr(x,abs(n),0); zeros(abs(n),cols(x))];
    end
end    

function z = rows(A)
    z = size(A,1);
end

function z = cols(A)
    z = size(A,2);
end

function seq = seqa(a,b,c)
% PURPOSE: produce a sequence of values
% -----------------------------------------------------
% USAGE: y = seqa(a,b,c)
%  where    a = initial value in sequence 
%           b = increment
%           c = number of values in the sequence  
% -----------------------------------------------------
% RETURNS: a sequence, (a:b:(a+b*(c-1)))' in MATLAB notation
% ----------------------------------------------------- 
% NOTE: a Gauss compatability function
% -----------------------------------------------------

% written by:
% James P. LeSage, Dept of Economics
% University of Toledo
% 2801 W. Bancroft St,
% Toledo, OH 43606
% jpl@jpl.econ.utoledo.edu
       
% seqa Gauss eqivalent of seqa(a,b,c)
seq = (a:b:(a+b*(c-1)))';
end

function [b, r, p] = olsqr2(y, A)
    [Q,R] = qr(A,0); 
    b = R\(Q'*y);
    r = y - A*b;
    p = A*b;
end

function z = trimr(x,n1,n2)
% PURPOSE: return a matrix (or vector) x stripped of the specified rows.
% -----------------------------------------------------
% USAGE: z = trimr(x,n1,n2)
% where: x = input matrix (or vector) (n x k)
%       n1 = first n1 rows to strip
%       n2 = last  n2 rows to strip
% NOTE: modeled after Gauss trimr function
% -----------------------------------------------------
% RETURNS: z = x(n1+1:n-n2,:)
% -----------------------------------------------------

% written by:
% James P. LeSage, Dept of Economics
% University of Toledo
% 2801 W. Bancroft St,
% Toledo, OH 43606
% jpl@jpl.econ.utoledo.edu

  [n, ~] = size(x);
  if (n1+n2) >= n; 
     error('Attempting to trim too much in trimr');
  end;
  h1 = n1+1;   
  h2 = n-n2;
  z = x(h1:h2,:);
end